@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014 Canonical Ltd.
+ * Copyright (C) 2014-2015 Canonical Ltd.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -38,6 +38,7 @@
 #include <ctype.h>
 #include <time.h>
 #include <getopt.h>
+#include <sched.h>
 
 #include <sys/ioctl.h>
 #include <sys/time.h>
@@ -69,6 +70,15 @@
 #define OPT_EV_MASK		(0x00001f00)
 #define OPT_EV_ALL		(OPT_EV_MASK)
 
+#define	GOT_TGID		(0x01)
+#define GOT_PPID		(0x02)
+#define GOT_ALL			(GOT_TGID | GOT_PPID)
+
+#ifndef LINUX_VERSION_CODE
+#define LINUX_VERSION_CODE KERNEL_VERSION(2,0,0)
+#endif
+
+
 /* /proc info cache */
 typedef struct proc_info {
 	pid_t	pid;		/* Process ID */
@@ -152,18 +162,9 @@ static const int signals[] = {
 #ifdef SIGQUIT
 	SIGQUIT,
 #endif
-#ifdef SIGILL
-	SIGILL,
-#endif
-#ifdef SIGABRT
-	SIGABRT,
-#endif
 #ifdef SIGFPE
 	SIGFPE,
 #endif
-#ifdef SIGSEGV
-	SIGSEGV,
-#endif
 #ifdef SIGTERM
 	SIGTERM,
 #endif
@@ -174,9 +175,6 @@ static const int signals[] = {
 	SIGUSR2,
 	/* POSIX.1-2001 */
 #endif
-#ifdef SIGBUS
-	SIGBUS,
-#endif
 #ifdef SIGXCPU
 	SIGXCPU,
 #endif
@@ -204,6 +202,57 @@ static const int signals[] = {
 
 static proc_info_t *proc_info_get(pid_t pid);
 
+/*
+ *  get_parent_pid()
+ *	get parent pid and set is_thread to true if process
+ *	not forked but a newly created thread
+ */
+static pid_t get_parent_pid(const pid_t pid, bool *is_thread)
+{
+	FILE *fp;
+	char path[PATH_MAX];
+	char buffer[4096];
+	pid_t tgid = 0, ppid = 0;
+	unsigned int got = 0;
+
+	*is_thread = false;
+	snprintf(path, sizeof(path), "/proc/%u/status", pid);
+	if ((fp = fopen(path, "r")) == NULL)
+		return 0;
+
+	while (((got & GOT_ALL) != GOT_ALL) &&
+	       (fgets(buffer, sizeof(buffer), fp) != NULL)) {
+		if (!strncmp(buffer, "Tgid:", 5)) {
+			if (sscanf(buffer + 5, "%u", &tgid) == 1) {
+				got |= GOT_TGID;
+			} else {
+				tgid = 0;
+			}
+		}
+		if (!strncmp(buffer, "PPid:", 5)) {
+			if (sscanf(buffer + 5, "%u", &ppid) == 1) {
+				got |= GOT_PPID;
+			} else {
+				ppid = 0;
+			}
+		}
+	}
+	fclose(fp);
+
+	if ((got & GOT_ALL) == GOT_ALL) {
+		/*  TGID and PID are not the same if it is a thread */
+		if (tgid != pid) {
+			/* In this case, the parent is the TGID */
+			ppid = tgid;
+			*is_thread = true;
+		}
+	} else {
+		ppid = 0;
+	}
+
+	return ppid;
+}
+
 /*
  *  sane_proc_pid_info()
  *	detect if proc info mapping from /proc/timer_stats
@@ -433,7 +482,7 @@ static void proc_stats_account(const pid_t pid, const event_t event)
  *  stats_cmp()
  *	compare stats total, used for sorting list
  */
-int stats_cmp(const void *v1, const void *v2)
+static int stats_cmp(const void *v1, const void *v2)
 {
 	proc_stats_t **s1 = (proc_stats_t **)v1;
 	proc_stats_t **s2 = (proc_stats_t **)v2;
@@ -448,7 +497,7 @@ int stats_cmp(const void *v1, const void *v2)
  *  proc_stats_report()
  *	report event statistics
  */
-void proc_stats_report(void)
+static void proc_stats_report(void)
 {
 	int i;
 	int n = 0;
@@ -497,7 +546,7 @@ void proc_stats_report(void)
  *  proc_stats_free()
  *	free stats list
  */
-void proc_stats_free(void)
+static void proc_stats_free(void)
 {
 	int i;
 
@@ -846,18 +895,47 @@ static int netlink_listen(const int sock)
 	return writev(sock, iov, 3);
 }
 
+/*
+ *   set_prioity
+ *	set high priority to try and get netlink activty
+ *	before short lived processes die
+ */
+static void set_priority(void)
+{
+	int max;
+	struct sched_param param;
+	int sched;
+
+#if defined(SCHED_DEADLINE)
+	sched = SCHED_DEADLINE;
+#elif defined(SCHED_SCHED_FIFO)
+	sched = SCHED_FIFO;
+#elif defined(SCHED_RR)
+	sched = SCHED_FIFO;
+#else
+	sched = SCHED_OTHER;	/* Oh well */
+#endif
+	if ((max = sched_get_priority_max(sched)) < 0)
+		return;
+
+	memset(&param, 0, sizeof(param));
+	param.sched_priority = max;
+	(void)sched_setscheduler(getpid(), sched, &param);
+}
+
 /*
  *   monitor()
  *	monitor system activity
  */
 static int monitor(const int sock)
 {
-	ssize_t len;
 	struct nlmsghdr *nlmsghdr;
 
 	print_heading();
+	set_priority();
 
 	while (!stop_recv) {
+		ssize_t len;
 		char __attribute__ ((aligned(NLMSG_ALIGNTO)))buf[4096];
 
 		if ((len = recv(sock, buf, sizeof(buf), 0)) == 0) {
@@ -900,12 +978,19 @@ static int monitor(const int sock)
 			struct cn_msg *cn_msg;
 			struct proc_event *proc_ev;
 			struct tm tm;
-			struct timeval tv;
-			time_t now;
 			char when[10];
+			time_t now;
+			pid_t ppid;
+			bool is_thread;
+
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
+			struct timeval tv;
 			char duration[32];
-			char *comm;
 			proc_info_t const *info1, *info2;
+#endif
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,1,0)
+			char *comm;
+#endif
 
 			if (stop_recv)
 				break;
@@ -937,22 +1022,27 @@ static int monitor(const int sock)
 				if (gettimeofday(&tv, NULL) < 0) {
 					memset(&tv, 0, sizeof tv);
 				}
-				info1 = proc_info_get(proc_ev->event_data.fork.parent_pid);
+				ppid = get_parent_pid(proc_ev->event_data.fork.child_pid, &is_thread);
+				info1 = proc_info_get(ppid);
 				info2 = proc_info_add(proc_ev->event_data.fork.child_pid, &tv);
 				if (!(opt_flags & OPT_QUIET) && (opt_flags & OPT_EV_FORK)) {
 					if (info1 != NULL && info2 != NULL) {
+						char *type = is_thread ? "clone" : "fork";
 						row_increment();
-						printf("%s fork %5d parent %8s %s%s%s\n",
+						printf("%s %-5.5s %5d parent %8s %s%s%s\n",
 							when,
-							proc_ev->event_data.fork.parent_pid,
+							type,
+							ppid,
 							"",
 							info1->kernel_thread ? "[" : "",
 							info1->cmdline,
 							info1->kernel_thread ? "]" : "");
 						row_increment();
-						printf("%s fork %5d child  %8s %s%s%s\n",
+						printf("%s %-5.5s %5d %-6.6s %8s %s%s%s\n",
 							when,
+							type,
 							proc_ev->event_data.fork.child_pid,
+							is_thread ? "thread" : "child",
 							"",
 							info1->kernel_thread ? "[" : "",
 							info2->cmdline,
@@ -965,7 +1055,7 @@ static int monitor(const int sock)
 				info1 = proc_info_update(proc_ev->event_data.exec.process_pid);
 				if (!(opt_flags & OPT_QUIET) && (opt_flags & OPT_EV_EXEC)) {
 					row_increment();
-					printf("%s exec %5d        %8s %s%s%s\n",
+					printf("%s exec  %5d        %8s %s%s%s\n",
 						when,
 						proc_ev->event_data.exec.process_pid,
 						"",
@@ -991,7 +1081,7 @@ static int monitor(const int sock)
 						snprintf(duration, sizeof(duration), "unknown");
 					}
 					row_increment();
-					printf("%s exit %5d  %5d %8s %s%s%s\n",
+					printf("%s exit  %5d  %5d %8s %s%s%s\n",
 						when,
 						proc_ev->event_data.exit.process_pid,
 						proc_ev->event_data.exit.exit_code,
@@ -1009,7 +1099,7 @@ static int monitor(const int sock)
 				if (!(opt_flags & OPT_QUIET) && (opt_flags & OPT_EV_CORE)) {
 					info1 = proc_info_get(proc_ev->event_data.coredump.process_pid);
 					row_increment();
-					printf("%s core %5d        %8s %s%s%s\n",
+					printf("%s core  %5d        %8s %s%s%s\n",
 						when,
 						proc_ev->event_data.exit.process_pid,
 						"",
@@ -1029,7 +1119,7 @@ static int monitor(const int sock)
 						break;
 					row_increment();
 
-					printf("%s comm %5d        %8s %s%s%s -> %s\n",
+					printf("%s comm  %5d        %8s %s%s%s -> %s\n",
 						when,
 						proc_ev->event_data.exit.process_pid,
 						"",
@@ -1053,7 +1143,7 @@ static int monitor(const int sock)
  *  show_help()
  *	simple help
  */
-void show_help(char *const argv[])
+static void show_help(char *const argv[])
 {
 	printf("%s, version %s\n\n", APP_NAME, VERSION);
 	printf("usage: %s [-d|-D|-e|-h|-s|-S|-q]\n", argv[0]);
